from __future__ import annotations

from typing import TYPE_CHECKING

__all__ = (
    "Alliance",
    "ApiKeyDetails",
    "Bankrec",
    "BBGame",
    "BBPlayer",
    "BBTeam",
    "Bounty",
    "City",
    "Color",
    "GameInfo",
    "Nation",
    "Trade",
    "Tradeprice",
    "Treasure",
    "Treaty",
    "War",
    "WarAttack",
    "PaginatorInfo",
)

if TYPE_CHECKING:
    from typing import Any, Dict, Mapping, Optional, Tuple, Type


class Data:
    id: str
    __slots__ = ()

    def __init__(self, data: Mapping[str, Any]) -> None:
        for key, value in data.items():
            if key == "global":
                key = "global_"
            if isinstance(value, dict):
                attr = _RELATION_MAP[key]
                object.__setattr__(self, key, attr(value))  # type: ignore
            elif isinstance(value, list):
                attr = _RELATION_MAP[key]
                object.__setattr__(self, key, tuple(attr(i) for i in value))  # type: ignore
            else:
                object.__setattr__(self, key, value)

    def __setattr__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delattr__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __setitem__(self, name: str, value: Any) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item assignment"
        )

    def __delitem__(self, name: str) -> None:
        raise TypeError(
            f"'{type(self).__name__}' object does not support item deletion"
        )

    def __getitem__(self, name: str) -> Any:
        try:
            return self.__getattribute__(name)
        except AttributeError:
            raise KeyError(name)

    def __int__(self) -> int:
        return int(self.get("id", -1))

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {int(self)}"

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self.__getattribute__(key)
        except AttributeError:
            return default

    def to_dict(self) -> Dict[str, Any]:
        """
        Get a dict representation of this data.

        Returns
        -------
        Dict[str, Any]
            A dictionary of the data on the object.
        """
        return {
            key: value
            for key in self.__slots__
            if (value := getattr(self, key, (False,))) != (False,)
        }


class Alliance(Data):
    """Represents an alliance.

    Attributes
    ----------
    id: :class:`str`
        The alliance ID.
    name: :class:`str`
        The alliance name.
    acronym: :class:`str`
        The alliance acronym.
    score: :class:`float`
        The alliance score.
    color: :class:`str`
        The alliance color.
    date: :class:`str`
        The date the alliance was created.
    nations: Tuple[:class:`Nation`, ...]
        The alliance's members.
    treaties: Tuple[:class:`Treaty`, ...]
        The alliance's treaties.
    acceptmem: :class:`bool`
        Whether the alliance is accepting members or not.
    flag: :class:`str`
        The alliance flag URL.
    forumlink: :class:`str`
        The alliance forum link.
    irclink: :class:`str`
        The alliance Discord link.
    bankrecs: Tuple[:class:`Bankrec`, ...]
        The alliance's bank records. Will not contain alliance to alliance records if the querying key cannot view the alliance bank.
    taxrecs: Optional[Tuple[:class:`Bankrec`, ...]]
        The alliance's tax records. Will be `None` if the querying key cannot view the alliance bank.
    tax_brackets: Optional[Tuple[:class:`TaxBracket`, ...]]
        The alliance's tax brackets. Will be `None` if the querying key cannot view the alliance bank.
    money: Optional[:class:`float`]
        The current amount of money in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    coal: Optional[:class:`float`]
        The current amount of coal in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    oil: Optional[:class:`float`]
        The current amount of oil in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    uranium: Optional[:class:`float`]
        The current amount of uranium in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    iron: Optional[:class:`float`]
        The current amount of iron in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    bauxite: Optional[:class:`float`]
        The current amount of bauxite in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    lead: Optional[:class:`float`]
        The current amount of lead in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    gasoline: Optional[:class:`float`]
        The current amount of gasoline in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    munitions: Optional[:class:`float`]
        The current amount of munitions in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    steel: Optional[:class:`float`]
        The current amount of steel in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    aluminum: Optional[:class:`float`]
        The current amount of aluminum in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    food: Optional[:class:`float`]
        The current amount of food in the alliance bank. Will be `None` if the querying key cannot view the alliance bank.
    """

    id: str
    name: str
    acronym: str
    score: float
    color: str
    date: str
    nations: Tuple[Nation, ...]
    treaties: Tuple[Treaty, ...]
    acceptmem: bool
    flag: str
    forumlink: str
    irclink: str
    bankrecs: Tuple[Bankrec, ...]
    taxrecs: Optional[Tuple[Bankrec, ...]]
    tax_brackets: Optional[Tuple[TaxBracket, ...]]
    money: Optional[float]
    coal: Optional[float]
    oil: Optional[float]
    uranium: Optional[float]
    iron: Optional[float]
    bauxite: Optional[float]
    lead: Optional[float]
    gasoline: Optional[float]
    munitions: Optional[float]
    steel: Optional[float]
    aluminum: Optional[float]
    food: Optional[float]

    # __slots__ = (
    #     "id",
    #     "name",
    #     "acronym",
    #     "score",
    #     "color",
    #     "date",
    #     "nations",
    #     "treaties",
    #     "acceptmem",
    #     "flag",
    #     "forumlink",
    #     "irclink",
    #     "bankrecs",
    #     "taxrecs",
    #     "tax_brackets",
    #     "money",
    #     "coal",
    #     "oil",
    #     "uranium",
    #     "iron",
    #     "bauxite",
    #     "lead",
    #     "gasoline",
    #     "munitions",
    #     "steel",
    #     "aluminum",
    #     "food",
    # )


class ApiKeyDetails(Data):
    nation: Nation


class Bankrec(Data):
    """Represents a bank record.

    Attributes
    ----------
    id: :class:`str`
        The bank record ID.
    date: :class:`str`
        The date the transaction was completed.
    sid: :class:`str`
        The ID of the sender.
    stype: :class:`int`
        The type of the sender. 1 corresponds to a :class:`Nation` and 2 corresponds to a :class:`Alliance`.
    rid: :class:`str`
        The ID of the receiver.
    rtype: :class:`int`
        The type of the receiver. 1 corresponds to a :class:`Nation` and 2 corresponds to a :class:`Alliance`
    pid: :class:`str`
        The nation ID of the banker.
    note: :class:`str`
        The transaction note.
    money: :class:`float`
        The money sent in the transaction.
    coal: :class:`float`
        The coal sent in the transaction.
    oil: :class:`float`
        The oil sent in the transaction.
    uranium: :class:`float`
        The uranium sent in the transaction.
    iron: :class:`float`
        The iron sent in the transaction.
    bauxite: :class:`float`
        The bauxite sent in the transaction.
    lead: :class:`float`
        The lead sent in the transaction.
    gasoline: :class:`float`
        The gasoline sent in the transaction.
    munitions: :class:`float`
        The munitions sent in the transaction.
    steel: :class:`float`
        The steel sent in the transaction.
    aluminum: :class:`float`
        The aluminum sent in the transaction.
    food: :class:`float`
        The food sent in the transaction.
    tax_id: :class:`str`
        The ID of the tax bracket.
    """

    id: str
    date: str
    sid: str
    stype: int
    rid: str
    rtype: int
    pid: str
    note: str
    money: float
    coal: float
    oil: float
    uranium: float
    iron: float
    bauxite: float
    lead: float
    gasoline: float
    munitions: float
    steel: float
    aluminum: float
    food: float
    tax_id: str

    # __slots__ = (
    #     "id",
    #     "date",
    #     "sid",
    #     "stype",
    #     "rid",
    #     "rtype",
    #     "pid",
    #     "note",
    #     "money",
    #     "coal",
    #     "oil",
    #     "uranium",
    #     "iron",
    #     "bauxite",
    #     "lead",
    #     "gasoline",
    #     "munitions",
    #     "steel",
    #     "aluminum",
    #     "food",
    #     "tax_id",
    # )


class BBGame(Data):
    """Represents a baseball game.

    Attributes
    ----------
    id: :class:`str`
        The game ID.
    date: :class:`str`
        The date of the game.
    home_id: :class:`str`
        The ID of the home team.
    away_id: :class:`str`
        The ID of the away team.
    home_team: :class:`BBTeam`
        The home team.
    away_team: :class:`BBTeam`
        The away team.
    home_nation_id: :class:`str`
        The ID of the home nation.
    away_nation_id: :class:`str`
        The ID of the away nation.
    home_nation: :class:`Nation`
        The home nation.
    away_nation: :class:`Nation`
        The away nation.
    stadium_name: :class:`str`
        The name of the stadium.
    home_score: :class:`int`
        The home team's score.
    away_score: :class:`int`
        The away team's score.
    sim_text: :class:`str`
        The simulation text.
    highlights: :class:`str`
        The game highlights.
    home_revenue: :class:`float`
        The home team's revenue.
    spoils: :class:`float`
        The spoils.
    open: :class:`int`
        Honestly no idea what this one does.
    wager: :class:`float`
        The wager.
    """

    id: str
    date: str
    home_id: str
    away_id: str
    home_team: BBTeam
    away_team: BBTeam
    home_nation_id: str
    away_nation_id: str
    home_nation: Nation
    away_nation: Nation
    stadium_name: str
    home_score: int
    away_score: int
    sim_text: str
    highlights: str
    home_revenue: float
    spoils: float
    open: int
    wager: float

    # __slots__ = (
    #     "id",
    #     "date",
    #     "home_id",
    #     "away_id",
    #     "home_team",
    #     "away_team",
    #     "home_nation_id",
    #     "away_nation_id",
    #     "home_nation",
    #     "away_nation",
    #     "stadium_name",
    #     "home_score",
    #     "away_score",
    #     "sim_text",
    #     "highlights",
    #     "home_revenue",
    #     "spoils",
    #     "open",
    #     "wager",
    # )


class BBPlayer(Data):
    """Represents a baseball player.

    Attributes
    ----------
    id: :class:`str`
        The player ID.
    date: :class:`str`
        The date the player was created.
    nation_id: :class:`str`
        The ID of the nation the player belongs to.
    nation: :class:`Nation`
        The nation the player belongs to.
    team_id: :class:`str`
        The ID of the team the player belongs to.
    team: :class:`BBTeam`
        The team the player belongs to.
    name: :class:`str`
        The player's name.
    age: :class:`int`
        The player's age.
    position: :class:`str`
        The player's position.
    pitching: :class:`float`
        The player's pitching rating.
    batting: :class:`float`
        The player's batting rating.
    speed: :class:`float`
        The player's speed rating.
    awareness: :class:`float`
        The player's awareness rating.
    overall: :class:`float`
        The player's overall rating.
    birthday: :class:`int`
        The player's birthday.
    """

    id: str
    date: str
    nation_id: str
    nation: Nation
    team_id: str
    team: BBTeam
    name: str
    age: int
    position: str
    pitching: float
    batting: float
    speed: float
    awareness: float
    overall: float
    birthday: int

    # __slots__ = (
    #     "id",
    #     "date",
    #     "nation_id",
    #     "nation",
    #     "team_id",
    #     "team",
    #     "name",
    #     "age",
    #     "position",
    #     "pitching",
    #     "batting",
    #     "speed",
    #     "awareness",
    #     "overall",
    #     "birthday",
    # )


class BBTeam(Data):
    """Represents a baseball team.

    Attributes
    ----------
    id: :class:`str`
        The team ID.
    date: :class:`str`
        The date the team was created.
    nation_id: :class:`str`
        The ID of the nation the team belongs to.
    nation: :class:`Nation`
        The nation the team belongs to.
    name: :class:`str`
        The team's name.
    logo: :class:`str`
        The team's logo.
    home_jersey: :class:`str`
        The team's home jersey.
    away_jersey: :class:`str`
        The team's away jersey.
    stadium: :class:`str`
        The team's stadium.
    quality: :class:`int`
        The team's quality rating.
    seating: :class:`int`
        The team's seating rating.
    rating: :class:`float`
        The team's overall rating.
    wins: :class:`int`
        The team's number of wins.
    glosses: :class:`int`
        The team's number of losses.
    runs: :class:`int`
        The team's number of runs.
    homers: :class:`int`
        The team's number of home runs.
    strikeouts: :class:`int`
        The team's number of strikeouts.
    games_played: :class:`int`
        The team's number of games played.
    games: Tuple[:class:`BBGame`, ...]
        The team's games.
    players: Tuple[:class:`BBPlayer`, ...]
        The team's players.
    """

    id: str
    date: str
    nation_id: str
    nation: Nation
    name: str
    logo: str
    home_jersey: str
    away_jersey: str
    stadium: str
    quality: int
    seating: int
    rating: float
    wins: int
    glosses: int
    runs: int
    homers: int
    strikeouts: int
    games_played: int
    games: Tuple[BBGame, ...]
    players: Tuple[BBPlayer, ...]

    # __slots__ = (
    #     "id",
    #     "date",
    #     "nation_id",
    #     "nation",
    #     "name",
    #     "logo",
    #     "home_jersey",
    #     "away_jersey",
    #     "stadium",
    #     "quality",
    #     "seating",
    #     "rating",
    #     "wins",
    #     "glosses",
    #     "runs",
    #     "homers",
    #     "strikeouts",
    #     "games_played",
    #     "games",
    #     "players",
    # )


class Bounty(Data):
    """Represents a bounty.

    Attributes
    ----------
    id: :class:`str`
        The bounty ID.
    date: :class:`str`
        The date the bounty was created.
    nation_id: :class:`str`
        The ID of the nation the bounty is on.
    nation: :class:`Nation`
        The nation the bounty is on.
    amount: :class:`int`
        The amount of the bounty reward.
    type: :class:`str`
        The type of bounty is for.
    """

    id: str
    date: str
    nation_id: str
    nation: Nation
    amount: int
    type: str

    # __slots__ = (
    #     "id",
    #     "date",
    #     "nation_id",
    #     "nation",
    #     "amount",
    #     "type",
    # )


class City(Data):
    """Represents a city.

    Attributes
    ----------
    id: :class:`str`
        The city ID.
    nation_id: :class:`str`
        The ID of the nation the city belongs to.
    nation: :class:`Nation`
        The nation the city belongs to.
    name: :class:`str`
        The city name.
    date: :class:`str`
        The date the city was created.
    infrastructure: :class:`float`
        The amount of infrastructure in the city.
    land: :class:`float`
        The amount of land in the city.
    powered: :class:`bool`
        Whether the city is powered or not.
    oilpower: :class:`int`
        The amount of Oil Power Plants in the city.
    windpower: :class:`int`
        The amount of Wind Power Plants in the city.
    coalpower: :class:`int`
        The amount of Coal Power Plants in the city.
    nuclearpower: :class:`int`
        The amount of Nuclear Power Plants in the city.
    coalmine: :class:`int`
        The amount of Coal Mines in the city.
    oilwell: :class:`int`
        The amount of Oil Wells in the city.
    uramine: :class:`int`
        The amount of Uranium Mines in the city.
    barracks: :class:`int`
        The amount of Barracks in the city.
    farm: :class:`int`
        The amount of Farms in the city.
    policestation: :class:`int`
        The amount of Police Stations in the city.
    hospital: :class:`int`
        The amount of Hospitals in the city.
    recyclingcenter: :class:`int`
        The amount of Recycling Centers in the city.
    subway: :class:`int`
        The amount of Subways in the city.
    supermarket: :class:`int`
        The amount of Supermarkets in the city.
    bank: :class:`int`
        The amount of Banks in the city.
    mall: :class:`int`
        The amount of Shopping Malls in the city.
    stadium: :class:`int`
        The amount of Stadiums in the city.
    leadmine: :class:`int`
        The amount of Lead Mines in the city.
    ironmine: :class:`int`
        The amount of Iron Mines in the city.
    bauxitemine: :class:`int`
        The amount of Bauxite Mines in the city.
    gasrefinery: :class:`int`
        The amount of Oil Refineries in the city.
    aluminumrefinery: :class:`int`
        The amount of Aluminum Refineries in the city.
    steelmill: :class:`int`
        The amount of Steel Mill in the city.
    munitionsfactory: :class:`int`
        The amount of Munitions Factories in the city.
    factory: :class:`int`
        The amount of Factories in the city.
    airforcebase: :class:`int`
        The amount of Hangars in the city.
    drydock: :class:`int`
        The amount of Drydocks in the city.
    nukedate: :class:`str`
        The date the city was last nuked.
    """

    id: str
    nation_id: str
    nation: Nation
    name: str
    date: str
    infrastructure: float
    land: float
    powered: bool
    oilpower: int
    windpower: int
    coalpower: int
    nuclearpower: int
    coalmine: int
    oilwell: int
    uramine: int
    barracks: int
    farm: int
    policestation: int
    hospital: int
    recyclingcenter: int
    subway: int
    supermarket: int
    bank: int
    mall: int
    stadium: int
    leadmine: int
    ironmine: int
    bauxitemine: int
    gasrefinery: int
    aluminumrefinery: int
    steelmill: int
    munitionsfactory: int
    factory: int
    airforcebase: int
    drydock: int
    nukedate: str

    # __slots__ = (
    #     "id",
    #     "nation_id",
    #     "nation",
    #     "name",
    #     "date",
    #     "infrastructure",
    #     "land",
    #     "powered",
    #     "oilpower",
    #     "windpower",
    #     "coalpower",
    #     "nuclearpower",
    #     "coalmine",
    #     "oilwell",
    #     "uramine",
    #     "barracks",
    #     "farm",
    #     "policestation",
    #     "hospital",
    #     "recyclingcenter",
    #     "subway",
    #     "supermarket",
    #     "bank",
    #     "mall",
    #     "stadium",
    #     "leadmine",
    #     "ironmine",
    #     "bauxitemine",
    #     "gasrefinery",
    #     "aluminumrefinery",
    #     "steelmill",
    #     "munitionsfactory",
    #     "factory",
    #     "airforcebase",
    #     "drydock",
    #     "nukedate",
    # )


class Color(Data):
    """Represents a color bloc.

    Attributes
    ----------
    color: :class:`str`
        The color.
    bloc_name: :class:`str`
        The color bloc name.
    turn_bonus: :class:`int`
        The color turn bonus.
    """

    color: str
    bloc_name: str
    turn_bonus: int

    # __slots__ = ("color", "bloc_name", "turn_bonus")

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {getattr(self, 'bloc_name', None)}"


class GameInfo(Data):
    """Represents some game info.

    Attributes
    ----------
    game_date: :class:`str`
        The game date.
    radiation: :class:`Radiation`
        The radiation index.
    """

    game_date: str
    radiation: Radiation

    # __slots__ = (
    #     "game_date",
    #     "radiation",
    # )


class Nation(Data):
    """Represents a nation.

    Attributes
    ----------
    id: :class:`str`
        The nation ID.
    alliance_id: :class:`str`
        The nation's alliance ID.
    alliance_position: :class:`str`
        The nation's alliance position.
    alliance: :class:`Alliance`
        The nation's alliance.
    nation_name: :class:`str`
        The nation name.
    leader_name: :class:`str`
        The nation's leader name.
    continent: :class:`str`
        The nation's continent.
    warpolicy: :class:`str`
        The nation's war policy.
    dompolicy: :class:`str`
        The nation's domestic policy.
    color: :class:`str`
        The nation's color
    num_cities: :class:`int`
        The number of cities the nation has.
    cities: Tuple[:class:`City`, ...]
        The nation's cities.
    score: :class:`float`
        The nation's score.
    update_tz: :class:`float`
        The nation's update timezone. Will return `None` if the querying key cannot see the nation's bank.
    population: :class:`int`
        The nation's population.
    flag: :class:`str`
        The nation's flag.
    vmode: :class:`int`
        The number of turns the nation has in Vacation Mode.
    beigeturns: :class:`int`
        The number of turns the nation has on Beige.
    espionage_available: :class:`bool`
        Whether the nation has can be have an espionage operation executed on it or not.
    last_active: :class:`str`
        When the nation was last active.
    date: :class:`str`
        When the nation was created.
    soldiers: :class:`int`
        The amount of soldiers the nation has.
    tanks: :class:`int`
        The amount of tanks the nation has.
    aircraft: :class:`int`
        The amount of aircraft the nation has.
    ships: :class:`int`
        The amount of ships the nation has.
    missiles: :class:`int`
        The amount of missiles the nation has.
    nukes: :class:`int`
        The amount of nukes the nation has.
    spies: :class:`int`
        The amount of spies the nation has. Will return `None` if the querying key cannot see the nation's bank.
    discord: :class:`str`
        The nation's discord username.
    treasures: Tuple[:class:`Treasure`, ...]
        The treasures a nation has.
    wars: Tuple[:class:`War`, ...]
        The wars the nation is involved in.
    bankrecs: Tuple[:class:`Bankrec`, ...]
        The received bank records of the nation.
    taxrecs: Optional[Tuple[:class:`Bankrec`, ...]]
        The nation's tax records. Will return `None` if the querying key cannot see the nation's bank.
    bounties: Tuple[:class:`Bounty`, ...]
        The bounties on the nation.
    turns_since_last_city: :class:`int`
        The number of turns since the nation has built a city.
    turns_since_last_project: :class:`int`
        The number of turns since the nation has built a project.
    money: Optional[:class:`float`]
        The amount of money the nation has. Will return `None` if the querying key cannot see the nation's bank.
    coal: Optional[:class:`float`]
        The amount of coal the nation has. Will return `None` if the querying key cannot see the nation's bank.
    oil: Optional[:class:`float`]
        The amount of oil the nation has. Will return `None` if the querying key cannot see the nation's bank.
    uranium: Optional[:class:`float`]
        The amount of uranium the nation has. Will return `None` if the querying key cannot see the nation's bank.
    iron: Optional[:class:`float`]
        The amount of iron the nation has. Will return `None` if the querying key cannot see the nation's bank.
    bauxite: Optional[:class:`float`]
        The amount of bauxite the nation has. Will return `None` if the querying key cannot see the nation's bank.
    lead: Optional[:class:`float`]
        The amount of lead the nation has. Will return `None` if the querying key cannot see the nation's bank.
    gasoline: Optional[:class:`float`]
        The amount of gasoline the nation has. Will return `None` if the querying key cannot see the nation's bank.
    munitions: Optional[:class:`float`]
        The amount of munitions the nation has. Will return `None` if the querying key cannot see the nation's bank.
    steel: Optional[:class:`float`]
        The amount of steel the nation has. Will return `None` if the querying key cannot see the nation's bank.
    aluminum: Optional[:class:`float`]
        The amount of aluminum the nation has. Will return `None` if the querying key cannot see the nation's bank.
    food: Optional[:class:`float`]
        The amount of food the nation has. Will return `None` if the querying key cannot see the nation's bank.
    projects: :class:`int`
        The number of projects the nation has.
    project_bits: :class:`int`
        The bitwise representation of the nation's projects.
    ironw: :class:`bool`
        Whether the nation has the Ironworks project or not.
    bauxitew: :class:`bool`
        Whether the nation has the Bauxiteworks project or not.
    armss: :class:`bool`
        Whether the nation has the Arms Stockpile project or not.
    egr: :class:`bool`
        Whether the nation has the Emergency Gasoline Reserve project or not.
    massirr: :class:`bool`
        Whether the nation has the Mass Irrigation project or not.
    itc: :class:`bool`
        Whether the nation has the International Trade Center project or not.
    mlp: :class:`bool`
        Whether the nation has the Missile Launch Pad project or not.
    nrf: :class:`bool`
        Whether the nation has the Nuclear Research Facility project or not.
    irond: :class:`bool`
        Whether the nation has the Iron Dome project or not.
    vds: :class:`bool`
        Whether the nation has the Vital Defense System project or not.
    cia: :class:`bool`
        Whether the nation has the Intelligence Agency project or not.
    cfce: :class:`bool`
        Whether the nation has the Center for Civil Engineering project or not.
    propb: :class:`bool`
        Whether the nation has the Propaganda Bureau project or not.
    uap: :class:`bool`
        Whether the nation has the Uranium Enrichment Program project or not.
    city_planning: :class:`bool`
        Whether the nation has the Urban Planning project or not.
    adv_city_planning: :class:`bool`
        Whether the nation has the Advanced Urban Planning project or not.
    space_program: :class:`bool`
        Whether the nation has the Space Program project or not.
    spy_satellite: :class:`bool`
        Whether the nation has the Spy Satellite project or not.
    moon_landing: :class:`bool`
        Whether the nation has the Moon Landing project or not.
    pirate_economy: :class:`bool`
        Whether the nation has the Pirate Economy project or not.
    recycling_initiative: :class:`bool`
        Whether the nation has the Recycling Initiative project or not.
    telecom_satellite: :class:`bool`
        Whether the nation has the Telecommunication Satellite project or not.
    green_tech: :class:`bool`
        Whether the nation has the Green Technologies project or not.
    arable_land_agency: :class:`bool`
        Whether the nation has the Arable Land Agency project or not.
    clinical_research_center: :class:`bool`
        Whether the nation has the Clinical Research Center project or not.
    specialized_police_training: :class:`bool`
        Whether the nation has the Specialized Police Training Program project or not.
    adv_engineering_corps: :class:`bool`
        Whether the nation has the Advanced Engineering Corps project or not.
    wars_won: :class:`int`
        The number of wars the nation has won.
    wars_lost: :class:`int`
        The number of wars the nation has lost.
    tax_id: :class:`str`
        The nation's tax bracket ID.
    alliance_seniority: :class:`int`
        The nation's alliance seniority.
    """

    id: str
    alliance_id: str
    alliance_position: str
    alliance: Alliance
    nation_name: str
    leader_name: str
    continent: str
    warpolicy: str
    dompolicy: str
    color: str
    num_cities: int
    cities: Tuple[City, ...]
    score: float
    update_tz: float
    population: int
    flag: str
    vmode: int
    beigeturns: int
    espionage_available: bool
    last_active: str
    date: str
    soldiers: int
    tanks: int
    aircraft: int
    ships: int
    missiles: int
    nukes: int
    spies: int
    discord: str
    treasures: Tuple[Treasure, ...]
    wars: Tuple[War, ...]
    bankrecs: Tuple[Bankrec, ...]
    taxrecs: Optional[Tuple[Bankrec, ...]]
    bounties: Tuple[Bounty, ...]
    turns_since_last_city: int
    turns_since_last_project: int
    money: Optional[float]
    coal: Optional[float]
    oil: Optional[float]
    uranium: Optional[float]
    iron: Optional[float]
    bauxite: Optional[float]
    lead: Optional[float]
    gasoline: Optional[float]
    munitions: Optional[float]
    steel: Optional[float]
    aluminum: Optional[float]
    food: Optional[float]
    projects: int
    project_bits: int
    ironw: bool
    bauxitew: bool
    armss: bool
    egr: bool
    massirr: bool
    itc: bool
    mlp: bool
    nrf: bool
    irond: bool
    vds: bool
    cia: bool
    cfce: bool
    propb: bool
    uap: bool
    city_planning: bool
    adv_city_planning: bool
    space_program: bool
    spy_satellite: bool
    moon_landing: bool
    pirate_economy: bool
    recycling_initiative: bool
    telecom_satellite: bool
    green_tech: bool
    arable_land_agency: bool
    clinical_research_center: bool
    specialized_police_training: bool
    adv_engineering_corps: bool
    wars_won: int
    wars_lost: int
    tax_id: str
    alliance_seniority: int

    # __slots__ = (
    #     "id",
    #     "alliance_id",
    #     "alliance_position",
    #     "alliance",
    #     "nation_name",
    #     "leader_name",
    #     "continent",
    #     "warpolicy",
    #     "dompolicy",
    #     "color",
    #     "num_cities",
    #     "cities",
    #     "score",
    #     "update_tz",
    #     "population",
    #     "flag",
    #     "vmode",
    #     "beigeturns",
    #     "espionage_available",
    #     "last_active",
    #     "date",
    #     "soldiers",
    #     "tanks",
    #     "aircraft",
    #     "ships",
    #     "missiles",
    #     "nukes",
    #     "spies",
    #     "discord",
    #     "treasures",
    #     "wars",
    #     "bankrecs",
    #     "taxrecs",
    #     "bounties",
    #     "turns_since_last_city",
    #     "turns_since_last_project",
    #     "money",
    #     "coal",
    #     "oil",
    #     "uranium",
    #     "iron",
    #     "bauxite",
    #     "lead",
    #     "gasoline",
    #     "munitions",
    #     "steel",
    #     "aluminum",
    #     "food",
    #     "projects",
    #     "project_bits",
    #     "ironw",
    #     "bauxitew",
    #     "armss",
    #     "egr",
    #     "massirr",
    #     "itc",
    #     "mlp",
    #     "nrf",
    #     "irond",
    #     "vds",
    #     "cia",
    #     "cfce",
    #     "propb",
    #     "uap",
    #     "city_planning",
    #     "adv_city_planning",
    #     "space_program",
    #     "spy_satellite",
    #     "moon_landing",
    #     "pirate_economy",
    #     "recycling_initiative",
    #     "telecom_satellite",
    #     "green_tech",
    #     "arable_land_agency",
    #     "clinical_research_center",
    #     "specialized_police_training",
    #     "adv_engineering_corps",
    #     "wars_won",
    #     "wars_lost",
    #     "tax_id",
    #     "alliance_seniority",
    # )


class Radiation(Data):
    """Represents the radiation index.

    Attributes
    ----------
    global_: :class:`float`
        The global radiation index.
    north_america: :class:`float`
        The North America radiation index.
    south_america: :class:`float`
        The South America radiation index.
    europe: :class:`float`
        The Europe radiation index.
    africa: :class:`float`
        The Africa radiation index.
    asia: :class:`float`
        The Asia radiation index.
    australia: :class:`float`
        The Australia radiation index.
    antarctica: :class:`float`
        The Antarctica radiation index.
    """

    global_: float
    north_america: float
    south_america: float
    europe: float
    africa: float
    asia: float
    australia: float
    antarctica: float

    # __slots__ = (
    #     "global_",
    #     "north_america",
    #     "south_america",
    #     "europe",
    #     "africa",
    #     "asia",
    #     "australia",
    #     "antarctica",
    # )


class TaxBracket(Data):
    """Represents a tax bracket.

    Attributes
    ----------
    id: :class:`str`
        The tax bracket ID.
    alliance_id: :class:`str`
        The tax bracket's alliance ID.
    alliance: :class:`Alliance`
        The tax bracket's alliance.
    date: :class:`str`
        The date the tax bracket was created.
    date_modified: :class:`str`
        The date the tax bracket was last modified.
    last_modifier_id: :class:`str`
        The ID of the last modifier of the tax bracket.
    last_modifier: :class:`Nation`
        The last modifier of the tax bracket.
    tax_rate: :class:`int`
        The tax rate of the tax bracket.
    resource_tax_rate: :class:`int`
        The resource tax rate of the tax bracket.
    bracket_name: :class:`str`
        The name of the tax bracket.
    """

    id: str
    alliance_id: str
    alliance: Alliance
    date: str
    date_modified: str
    last_modifier_id: str
    last_modifier: Nation
    tax_rate: int
    resource_tax_rate: int
    bracket_name: str

    # __slots__ = (
    #     "id",
    #     "alliance_id",
    #     "alliance",
    #     "date",
    #     "date_modified",
    #     "last_modifier_id",
    #     "last_modifier",
    #     "tax_rate",
    #     "resource_tax_rate",
    #     "bracket_name",
    # )


class Trade(Data):
    """Represents a trade

    Attributes
    ----------
    id: :class:`str`
        The trade ID.
    type: :class:`str`
        The type of trade.
    date: :class:`str`
        The date the trade took place.
    sid: :class:`str`
        The sender ID.
    rid: :class:`str`
        The receiver ID.
    sender: :class:`Nation`
        The sender's nation.
    receiver: :class:`Nation`
        The receiver's nation.
    offer_resource: :class:`str`
        The resource being offered.
    offer_amount: :class:`int`
        The amount being offered.
    buy_or_sell: :class:`str`
        Whether the offer was a `buy` or `sell` offer.
    total: :class:`int`
        The price per unit of the offer.
    accepted: :class:`bool`
        Whether or not the offer has been accepted.
    date_accepted: :class:`str`
        The date the offer was accepted.
    """

    id: str
    type: str
    date: str
    sid: str
    rid: str
    sender: Nation
    receiver: Nation
    offer_resource: str
    offer_amount: int
    buy_or_sell: str
    total: int
    accepted: bool
    date_accepted: str

    # __slots__ = (
    #     "id",
    #     "type",
    #     "date",
    #     "sid",
    #     "rid",
    #     "sender",
    #     "receiver",
    #     "offer_resource",
    #     "offer_amount",
    #     "buy_or_sell",
    #     "total",
    #     "accepted",
    #     "date_accepted",
    # )


class Tradeprice(Data):
    """Represents a tradeprice.

    Attributes
    ----------
    id: :class:`str`
        The tradeprice ID.
    date: :class:`str`
        The date of the tradeprice.
    coal: :class:`float`
        The price of coal.
    oil: :class:`float`
        The price of oil.
    uranium: :class:`float`
        The price of uranium.
    iron: :class:`float`
        The price of iron.
    bauxite: :class:`float`
        The price of bauxite.
    lead: :class:`float`
        The price of lead.
    gasoline: :class:`float`
        The price of gasoline.
    munitions: :class:`float`
        The price of munitions.
    steel: :class:`float`
        The price of steel.
    aluminum: :class:`float`
        The price of aluminum.
    food: :class:`float`
        The price of food.
    credits: :class:`float`
        The price of credits.
    """

    id: str
    date: str
    coal: float
    oil: float
    uranium: float
    iron: float
    bauxite: float
    lead: float
    gasoline: float
    munitions: float
    steel: float
    aluminum: float
    food: float
    credits: float

    # __slots__ = (
    #     "id",
    #     "date",
    #     "coal",
    #     "oil",
    #     "uranium",
    #     "iron",
    #     "bauxite",
    #     "lead",
    #     "gasoline",
    #     "munitions",
    #     "steel",
    #     "aluminum",
    #     "food",
    #     "credits",
    # )


class Treasure(Data):
    """Represents a treasure.

    Attributes
    ----------
    name: :class:`str`
        The name of the treasure.
    color: :class:`str`
        The treasure's color.
    continent: :class:`str`
        The treasure's continent.
    bonus: :class:`int`
        The treasure's bonus.
    spawndate: :class:`str`
        The treasure's spawndate.
    nation: :class:`Nation`
        The nation that has the treasure.
    """

    name: str
    color: str
    continent: str
    bonus: int
    spawndate: str
    nation: Nation

    # __slots__ = (
    #     "name",
    #     "color",
    #     "continent",
    #     "bonus",
    #     "spawndate",
    #     "nation",
    # )

    def __repr__(self) -> str:
        return f"{type(self).__name__} - {getattr(self, 'name', None)}"


class Treaty(Data):
    """Represents a treaty.

    Attributes
    ----------
    id: :class:`str`
        The treaty ID.
    date: :class:`str`
        The date and time the treaty was created.
    treaty_type: :class:`str`
        The type of treaty.
    turns_left: :class:`int`
        The number of turns left in the treaty.
    alliance1_id: :class:`str`
        The ID of the alliance who sent the treaty.
    alliance1: :class:`Alliance`
        The alliance who sent the treaty.
    alliance2_id: :class:`str`
        The ID of the alliance who received the treaty.
    alliance2: :class:`Alliance`
        The alliance who received the treaty.
    """

    date: str
    treaty_type: str
    turns_left: int
    alliance1_id: str
    alliance1: Alliance
    alliance2_id: str
    alliance2: Alliance

    # __slots__ = (
    #     "id",
    #     "date",
    #     "treaty_type",
    #     "turns_left",
    #     "alliance1_id",
    #     "alliance1",
    #     "alliance2_id",
    #     "alliance2",
    # )


class War(Data):
    """Represents a war.

    Attributes
    ----------
    id: :class:`str`
        The war ID.
    date: :class:`str`
        The date the war was declared.
    reason: :class:`str`
        The war reason.
    war_type: :class:`str`
        The war type.
    groundcontrol: :class:`str`
        The ID of the nation that has Ground Control.
    airsuperiority: :class:`str`
        The ID of the nation that has Air Superiority.
    navalblockade: :class:`str`
        The ID of the nation that has a Naval Blockade.
    winner: :class:`str`
        The ID of the winner of the war.
    attacks: Tuple[:class:`WarAttack`, ...]
        The attacks in the war.
    turnsleft: :class:`int`
        The turns left in the war.
    attid: :class:`str`
        The attacker ID.
    att_alliance_id: :class:`str`
        The attacker's alliance ID.
    attacker: :class:`Nation`
        The attacker's nation.
    defid: :class:`str`
        The defender's ID.
    def_alliance_id: :class:`str`
        The defender's alliance ID.
    defender: :class:`Nation`
        The defender's nation.
    attpoints: :class:`int`
        The attacker's Military Action Points.
    defpoints: :class:`int`
        The defender's Military Action Points.
    attpeace: :class:`bool`
        Whether the attacker has offered peace or not.
    defpeace: :class:`bool`
        Whether the defender has offered peace or not.
    att_resistance: :class:`int`
        The attacker's resistance.
    def_resistance: :class:`int`
        The defender's resistance.
    att_fortify: :class:`bool`
        Whether the attacker has fortified or not.
    def_fortify: :class:`bool`
        Whether the defender has fortified or not.
    att_gas_used: :class:`float`
        The amount of gasoline used by the attacker.
    def_gas_used: :class:`float`
        The amount of gasoline used by the defender.
    att_mun_used: :class:`float`
        The amount of munitions used by the attacker.
    def_mun_used: :class:`float`
        The amount of munitions used by the defender.
    att_alum_used: :class:`int`
        The amount of aluminum used by the attacker.
    def_alum_used: :class:`int`
        The amount of aluminum used by the defender.
    att_steel_used: :class:`int`
        The amount of steel used by the attacker.
    def_steel_used: :class:`int`
        The amount of steel used by the defender.
    att_infra_destroyed: :class:`float`
        The amount of infrastructure destroyed by the attacker.
    def_infra_destroyed: :class:`float`
        The amount of infrastructure destroyed by the defender.
    att_money_looted: :class:`float`
        The amount of money looted by the attacker.
    def_money_looted: :class:`float`
        The amount of money looted by the attacker.
    att_soldiers_killed: :class:`int`
        The amount of soldiers killed by the attacker.
    def_soldiers_killed: :class:`int`
        The amount of soldiers killed by the defender.
    att_tanks_killed: :class:`int`
        The amount of tanks killed by the attacker.
    def_tanks_killed: :class:`int`
        The amount of tanks killed by the defender.
    att_aircraft_killed: :class:`int`
        The amount of aircraft killed by the attacker.
    def_aircraft_killed: :class:`int`
        The amount of aircraft killed by the defender.
    att_ships_killed: :class:`int`
        The amount of ships killed by the attacker.
    def_ships_killed: :class:`int`
        The amount of ships killed by the defender.
    att_missiles_used: :class:`int`
        The amount of missiles used by the attacker.
    def_missiles_used: :class:`int`
        The amount of missiles killed by the defender.
    att_nukes_used: :class:`int`
        The amount of nukes used by the attacker.
    def_nukes_used: :class:`int`
        The amount of nukes killed by the defender.
    att_infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed by the attacker.
    def_infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed by the defender.
    """

    id: str
    date: str
    reason: str
    war_type: str
    groundcontrol: str
    airsuperiority: str
    navalblockade: str
    winner: str
    attacks: Tuple[WarAttack, ...]
    turnsleft: int
    attid: str
    att_alliance_id: str
    attacker: Nation
    defid: str
    def_alliance_id: str
    defender: Nation
    attpoints: int
    defpoints: int
    attpeace: bool
    defpeace: bool
    att_resistance: int
    def_resistance: int
    att_fortify: bool
    def_fortify: bool
    att_gas_used: float
    def_gas_used: float
    att_mun_used: float
    def_mun_used: float
    att_alum_used: int
    def_alum_used: int
    att_steel_used: int
    def_steel_used: int
    att_infra_destroyed: float
    def_infra_destroyed: float
    att_money_looted: float
    def_money_looted: float
    att_soldiers_killed: int
    def_soldiers_killed: int
    att_tanks_killed: int
    def_tanks_killed: int
    att_aircraft_killed: int
    def_aircraft_killed: int
    att_ships_killed: int
    def_ships_killed: int
    att_missiles_used: int
    def_missiles_used: int
    att_nukes_used: int
    def_nukes_used: int
    att_infra_destroyed_value: float
    def_infra_destroyed_value: float

    # __slots__ = (
    #     "id",
    #     "date",
    #     "reason",
    #     "war_type",
    #     "groundcontrol",
    #     "airsuperiority",
    #     "navalblockade",
    #     "winner",
    #     "attacks",
    #     "turnsleft",
    #     "attid",
    #     "att_alliance_id",
    #     "attacker",
    #     "defid",
    #     "def_alliance_id",
    #     "defender",
    #     "attpoints",
    #     "defpoints",
    #     "attpeace",
    #     "defpeace",
    #     "att_resistance",
    #     "def_resistance",
    #     "att_fortify",
    #     "def_fortify",
    #     "att_gas_used",
    #     "def_gas_used",
    #     "att_mun_used",
    #     "def_mun_used",
    #     "att_alum_used",
    #     "def_alum_used",
    #     "att_steel_used",
    #     "def_steel_used",
    #     "att_infra_destroyed",
    #     "def_infra_destroyed",
    #     "att_money_looted",
    #     "def_money_looted",
    #     "att_soldiers_killed",
    #     "def_soldiers_killed",
    #     "att_tanks_killed",
    #     "def_tanks_killed",
    #     "att_aircraft_killed",
    #     "def_aircraft_killed",
    #     "att_ships_killed",
    #     "def_ships_killed",
    #     "att_missiles_used",
    #     "def_missiles_used",
    #     "att_nukes_used",
    #     "def_nukes_used",
    #     "att_infra_destroyed_value",
    #     "def_infra_destroyed_value",
    # )


class WarAttack(Data):
    """Represents a war attack.

    Attributes
    ----------
    id: :class:`str`
        The attack ID.
    date: :class:`str`
        The date the attack took place.
    attid: :class:`str`
        The attacking nation ID.
    attacker: :class:`Nation`
        The attacking nation.
    defid: :class:`str`
        The defending nation ID.
    defender: :class:`Nation`
        The defending nation.
    type: :class:`str`
        The attack type.
    warid: :class:`str`
        The war ID.
    war: :class:`War`
        The war.
    victor: :class:`str`
        The victor of the attack.
    success: :class:`int`
        The attack's level of success.
    attcas1: :class:`int`
        A special value depending on the attack.
    defcas1: :class:`int`
        A special value depending on the attack.
    attcas2: :class:`int`
        A special value depending on the attack.
    defcas2: :class:`int`
        A special value depending on the attack.
    cityid: :class:`str`
        The ID of the city that was attacked.
    infradestroyed: :class:`float`
        The amount of infrastructure destroyed.
    improvementslost: :class:`int`
        The amount of improvements destroyed.
    moneystolen: :class:`float`
        The amount of money stolen.
    loot_info: :class:`str`
        The loot info of the attack.
    resistance_eliminated: :class:`int`
        The amount of resistance eliminated.
    city_infra_before: :class:`float`
        The city's infrastructure before the attack
    infra_destroyed_value: :class:`float`
        The value of infrastructure destroyed.
    att_mun_used: :class:`float`
        The amount of munitions used by the attacker.
    def_mun_used: :class:`float`
        The amount of munitions used by the defender.
    att_gas_used: :class:`float`
        The amount of gasoline used by the attacker.
    def_gas_used: :class:`float`
        The amount of gasoline used by the defender.
    aircraft_killed_by+tanks: :class:`int`
        The amount of aircraft killed by tanks.
    """

    id: str
    date: str
    attid: str
    attacker: Nation
    defid: str
    defender: Nation
    type: str
    warid: str
    war: War
    victor: str
    success: int
    attcas1: int
    defcas1: int
    attcas2: int
    defcas2: int
    cityid: str
    infradestroyed: float
    improvementslost: int
    moneystolen: float
    loot_info: str
    resistance_eliminated: int
    city_infra_before: float
    infra_destroyed_value: float
    att_mun_used: float
    def_mun_used: float
    att_gas_used: float
    def_gas_used: float
    aircraft_killed_by_tanks: int

    # __slots__ = (
    #     "id",
    #     "date",
    #     "attid",
    #     "attacker",
    #     "defid",
    #     "defender",
    #     "type",
    #     "warid",
    #     "war",
    #     "victor",
    #     "success",
    #     "attcas1",
    #     "defcas1",
    #     "attcas2",
    #     "defcas2",
    #     "cityid",
    #     "infradestroyed",
    #     "improvementslost",
    #     "moneystolen",
    #     "loot_info",
    #     "resistance_eliminated",
    #     "city_infra_before",
    #     "infra_destroyed_value",
    #     "att_mun_used",
    #     "def_mun_used",
    #     "att_gas_used",
    #     "def_gas_used",
    #     "aircraft_killed_by_tanks",
    # )


class PaginatorInfo(Data):
    """Represents paginator info

    Attributes
    ----------
    count: :class:`int`
        Count of items available on the page.
    currentPage: :class:`int`
        The current page number.
    firstItem: :class:`int`
        The index of the first item on the page.
    hasMorePages: :class:`bool`
        Whether there are more pages or not.
    lastItem: :class:`int`
        The index of the last item on the page.
    lastPage: :class:`int`
        The page number of the last page.
    perPage: :class:`int`
        The number of items per page.
    total: :class:`int`
        The total number of items.
    """

    count: int
    currentPage: int
    firstItem: int
    hasMorePages: bool
    lastItem: int
    lastPage: int
    perPage: int
    total: int

    # __slots__ = (
    #     "count",
    #     "currentPage",
    #     "firstItem",
    #     "hasMorePages",
    #     "lastItem",
    #     "lastPage",
    #     "perPage",
    #     "total",
    # )


_RELATION_MAP: Mapping[str, Type[Data]] = {
    "alliance": Alliance,
    "cities": City,
    "treasures": Treasure,
    "attacks": WarAttack,
    "attacker": Nation,
    "defender": Nation,
    "sender": Nation,
    "receiver": Nation,
    "nations": Nation,
    "nation": Nation,
    "bankrecs": Bankrec,
    "taxrecs": Bankrec,
    "offensive_wars": War,
    "defensive_wars": War,
    "wars": War,
    "alliance1": Alliance,
    "alliance2": Alliance,
    "tax_brackets": TaxBracket,
    "treaties": Treaty,
    "home_team": BBTeam,
    "away_team": BBTeam,
    "home_nation": Nation,
    "away_nation": Nation,
    "team": BBTeam,
    "games": BBGame,
    "players": BBPlayer,
    "last_modifier": Alliance,
    "bounties": Bounty,
    "war": War,
}



"""
The MIT License (MIT)

Copyright (c) 2021-present Village

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
import sys
import traceback
from typing import TYPE_CHECKING

from . import data

__all__ = (
    "get_int_or_none",
    "get_datetime_or_none",
    "convert_data_array",
    "convert_data_dict",
    "find_data_class",
    "find_event_data_class",
    "print_exception_with_header",
    "print_exception",
)

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Union


def get_int_or_none(value: Optional[Any]) -> Optional[int]:
    return None if value is None else int(value)


def get_datetime_or_none(value: Optional[Any]) -> Optional[datetime.datetime]:
    return None if value is None else datetime.datetime.fromisoformat(value)


def convert_data_array(data: List[Any]) -> List[Any]:
    if not data:
        return []
    cls = find_data_class(data[0]["__typename"])
    return [cls.from_data(i) for i in data]


def convert_data_dict(data: Dict[Any, Any]) -> Union[Any, List[Any]]:
    if data["__typename"].endswith("Paginator"):
        return convert_data_array(data["data"])
    return find_data_class(data["__typename"]).from_data(data)


def find_data_class(name: str) -> Any:
    return getattr(data, name)


def find_event_data_class(name: str) -> Any:
    name = remove_prefix(name, "BULK_")
    name = remove_suffix(name, "_CREATE", "_UPDATE", "_DELETE")
    name = "".join(i.capitalize() for i in name.split("_"))
    if name.lower() == "warattack":
        name = "WarAttack"
    return getattr(data, name)


def print_exception_with_header(header: str, error: Exception) -> None:
    print(header, file=sys.stderr)
    print_exception(error)


def print_exception(error: Exception) -> None:
    traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)
    sys.stderr.flush()


def remove_prefix(s: str, *prefixes: str) -> str:
    for prefix in prefixes:
        if s.startswith(prefix):
            return s[len(prefix) :]
    return s


def remove_suffix(s: str, *suffixes: str) -> str:
    for suffix in suffixes:
        if s.endswith(suffix):
            return s[: -len(suffix)]
    return s



from __future__ import annotations

from typing import TYPE_CHECKING, overload

import requests

from ..errors import GraphQLError
from .base import KitBase
from .data import (
    Alliance,
    ApiKeyDetails,
    Bankrec,
    BBGame,
    BBPlayer,
    BBTeam,
    Bounty,
    City,
    Color,
    Data,
    GameInfo,
    Nation,
    Trade,
    Tradeprice,
    Treasure,
    Treaty,
    War,
    WarAttack,
)
from .paginator import Paginator

if TYPE_CHECKING:
    from typing import (
        Any,
        Dict,
        Final,
        Literal,
        Mapping,
        MutableMapping,
        Optional,
        Sequence,
        Tuple,
        Type,
        TypeVar,
        Union,
    )

    D = TypeVar("D", bound=Data)


class SyncKit(KitBase):
    is_async: Final[bool] = False

    def actual_query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        args: Sequence[Union[str, Any]],
        variables: MutableMapping[str, Any],
        is_paginator: bool = False,
        root: str = "query",
        headers: Optional[MutableMapping[str, Any]] = None,
    ) -> Dict[str, Any]:
        variables, variables_string = self._format_variables(variables)
        query = f"{root}{f'({variables_string})' if variables_string else ''}{self._format_query(endpoint, params, args, is_paginator)}"
        with requests.request(
            "POST",
            self.graphql_url,
            json={"query": query, "variables": variables},
            headers=headers,
        ) as response:
            data: Any = response.json(
                parse_int=self.parse_int, parse_float=self.parse_float
            )
            try:
                if "errors" in data[0]:
                    error = (
                        "\n".join(i["message"] for i in data[0]["errors"])
                        if len(data[0]["errors"]) > 1
                        else data[0]["errors"][0]["message"]
                    )
                    raise GraphQLError(error)
            except KeyError:
                pass
            try:
                if "errors" in data:
                    error = (
                        "\n".join(i["message"] for i in data["errors"])
                        if len(data["errors"]) > 1
                        else data["errors"][0]["message"]
                    )
                    raise GraphQLError(error)
            except KeyError:
                pass
            return data

    def run_query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        args: Sequence[Union[str, Any]],
        variables: MutableMapping[str, Any],
        type: Type[D],
        is_paginator: bool = False,
        root: str = "query",
        headers: Optional[MutableMapping[str, Any]] = None,
    ) -> Tuple[D, ...]:
        data = self.actual_query(
            endpoint, params, args, variables, is_paginator, root, headers
        )
        if is_paginator:
            return tuple(type(i) for i in data["data"][endpoint]["data"])
        return tuple(type(i) for i in data["data"][endpoint])

    @overload
    def query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False],
        is_paginator: bool,
        type: Type[D],
        root: str = ...,
        headers: Optional[MutableMapping[str, Any]] = ...,
        **variables: Any,
    ) -> Tuple[D, ...]:
        ...

    @overload
    def query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True],
        is_paginator: bool,
        type: Type[D],
        root: str = ...,
        headers: Optional[MutableMapping[str, Any]] = ...,
        **variables: Any,
    ) -> Paginator[D]:
        ...

    @overload
    def query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool,
        is_paginator: bool,
        type: Type[D],
        root: str = ...,
        headers: Optional[MutableMapping[str, Any]] = ...,
        **variables: Any,
    ) -> Union[Tuple[D, ...], Paginator[D]]:
        ...

    def query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool,
        is_paginator: bool,
        type: Type[D],
        root: str = "query",
        headers: Optional[MutableMapping[str, Any]] = None,
        **variables: Any,
    ) -> Union[Tuple[D, ...], Paginator[D]]:
        if paginator:
            return Paginator(None, endpoint, self, params, args, variables, type)
        return self.run_query(
            endpoint, params, args, variables, type, is_paginator, root, headers
        )

    def single_query(
        self,
        endpoint: str,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        type: Type[D],
        root: str = "query",
        headers: Optional[MutableMapping[str, Any]] = None,
        **variables: Any,
    ) -> D:
        data = self.actual_query(
            endpoint, params, args, variables, False, root, headers
        )
        return type(data["data"][endpoint])

    @overload
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Alliance, ...]:
        ...

    @overload
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Alliance]:
        ...

    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Alliance, ...], Paginator[Alliance]]:
        return self.query(
            "alliances",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Alliance,
            **variables,
        )

    @overload
    def bankrec_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Bankrec, ...]:
        ...

    @overload
    def bankrec_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Bankrec]:
        ...

    def bankrec_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Bankrec, ...], Paginator[Bankrec]]:
        return self.query(
            "bankrecs",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Bankrec,
            **variables,
        )

    @overload
    def baseball_game_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[BBGame, ...]:
        ...

    @overload
    def baseball_game_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[BBGame]:
        ...

    def baseball_game_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[BBGame, ...], Paginator[BBGame]]:
        return self.query(
            "baseball_games",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=BBGame,
            **variables,
        )

    @overload
    def baseball_player_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[BBPlayer, ...]:
        ...

    @overload
    def baseball_player_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[BBPlayer]:
        ...

    def baseball_player_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[BBPlayer, ...], Paginator[BBPlayer]]:
        return self.query(
            "baseball_players",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=BBPlayer,
            **variables,
        )

    @overload
    def baseball_team_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[BBTeam, ...]:
        ...

    @overload
    def baseball_team_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[BBTeam]:
        ...

    def baseball_team_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[BBTeam, ...], Paginator[BBTeam]]:
        return self.query(
            "baseball_teams",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=BBTeam,
            **variables,
        )

    @overload
    def bounty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Bounty, ...]:
        ...

    @overload
    def bounty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Bounty]:
        ...

    def bounty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Bounty, ...], Paginator[Bounty]]:
        return self.query(
            "bounties",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Bounty,
            **variables,
        )

    @overload
    def city_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[City, ...]:
        ...

    @overload
    def city_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[City]:
        ...

    def city_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[City, ...], Paginator[City]]:
        return self.query(
            "cities",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=City,
            **variables,
        )

    def color_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
    ) -> Tuple[Color, ...]:
        return self.query(
            "colors",
            params,
            *args,
            paginator=False,
            is_paginator=False,
            type=Color,
        )

    def game_info_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
    ) -> GameInfo:
        return self.single_query(
            "game_info",
            params,
            *args,
            paginator=False,
            is_paginator=False,
            type=GameInfo,
        )

    def me_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
    ) -> ApiKeyDetails:
        ...

    @overload
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Nation, ...]:
        ...

    @overload
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Nation]:
        ...

    def nation_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Nation, ...], Paginator[Nation]]:
        return self.query(
            "nations",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Nation,
            **variables,
        )

    @overload
    def trade_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Trade, ...]:
        ...

    @overload
    def trade_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Trade]:
        ...

    def trade_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Trade, ...], Paginator[Trade]]:
        return self.query(
            "trades",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Trade,
            **variables,
        )

    @overload
    def tradeprice_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Tradeprice, ...]:
        ...

    @overload
    def tradeprice_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Tradeprice]:
        ...

    def tradeprice_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Tradeprice, ...], Paginator[Tradeprice]]:
        return self.query(
            "tradeprices",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Tradeprice,
            **variables,
        )

    def treasure_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
    ) -> Tuple[Treasure, ...]:
        return self.query(
            "treasures",
            params,
            *args,
            paginator=False,
            is_paginator=False,
            type=Treasure,
        )

    @overload
    def treaty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[Treaty, ...]:
        ...

    @overload
    def treaty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Treaty]:
        ...

    def treaty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[Treaty, ...], Paginator[Treaty]]:
        return self.query(
            "treaties",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=Treaty,
            **variables,
        )

    @overload
    def war_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[War, ...]:
        ...

    @overload
    def war_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[War]:
        ...

    def war_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[War, ...], Paginator[War]]:
        return self.query(
            "wars",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=War,
            **variables,
        )

    @overload
    def warattack_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Tuple[WarAttack, ...]:
        ...

    @overload
    def warattack_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[WarAttack]:
        ...

    def warattack_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Tuple[WarAttack, ...], Paginator[WarAttack]]:
        return self.query(
            "warattacks",
            params,
            *args,
            paginator=paginator,
            is_paginator=True,
            type=WarAttack,
            **variables,
        )

    def bank_deposit_mutation(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        **variables: Any,
    ) -> Bankrec:
        return self.single_query(
            "bankDeposit",
            params,
            *args,
            type=Bankrec,
            root="mutation",
            headers={"X-Bot-Key": self.bot_key, "X-Api-Key": self.bot_key_api_key},
            **variables,
        )

    def bank_withdraw_mutation(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        **variables: Any,
    ) -> Bankrec:
        return self.single_query(
            "bankWithdraw",
            params,
            *args,
            type=Bankrec,
            root="mutation",
            headers={"X-Bot-Key": self.bot_key, "X-Api-Key": self.bot_key_api_key},
            **variables,
        )



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from . import utils

__all__ = ("RATE_LIMITS", "RateLimit")

if TYPE_CHECKING:
    from typing import Any, Dict, Optional

RATE_LIMITS: Dict[str, RateLimit] = {}


class RateLimit:
    def __init__(self, url: str) -> None:
        self.url: str = url
        self.limit: Optional[int] = None
        self.remaining: Optional[int] = None
        self.reset: Optional[int] = None
        self.interval: Optional[int] = None

    @property
    def initialized(self) -> bool:
        return (
            self.limit is not None
            and self.remaining is not None
            and self.reset is not None
            and self.interval is not None
        )

    def initialize(self, headers: Any) -> None:
        self.limit = utils.get_int_or_none(headers.get("X-RateLimit-Limit"))
        self.remaining = utils.get_int_or_none(headers.get("X-RateLimit-Remaining"))
        self.reset = utils.get_int_or_none(headers.get("X-RateLimit-Reset"))
        self.interval = utils.get_int_or_none(headers.get("X-RateLimit-Interval"))

    def hit(self) -> int:
        if (
            self.limit is None
            or self.remaining is None
            or self.reset is None
            or self.interval is None
        ):
            return 0
        current_time = int(time.time())
        if current_time > self.reset:
            self.remaining = self.limit - 1
            self.reset = int(current_time + 1 + self.interval)
            return 0
        self.remaining -= 1
        if self.remaining <= 0:
            return int(self.reset - current_time) + 1
        return 0

    @classmethod
    def get(cls, url: str) -> RateLimit:
        if url not in RATE_LIMITS:
            RATE_LIMITS[url] = cls(url)
        return RATE_LIMITS[url]

    def handle_429(self, reset: Optional[str]) -> Optional[float]:
        self.remaining = 0
        self.reset = (
            utils.get_int_or_none(reset)
            or self.reset
            or int(time.time() + (self.interval or 60))
        )
        return self.interval if self.reset is None else self.reset - time.time()



from __future__ import annotations

from csv import DictReader, reader
from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple

import aiohttp

from .abc import DataBase
from .data import _TYPE_MAP, DumpData

if TYPE_CHECKING:
    from csv import Dialect
    from typing import Iterator, List

    class _reader(Iterator[List[str]]):
        dialect: Dialect
        line_num: int

        def __next__(self) -> List[str]:
            ...


# when parsing bytes as string the 0th byte is taken
# off since it's has no place in the string
class AsyncData(DataBase):
    @classmethod
    async def _request(cls, type_: str, date: str) -> bytes:
        type_, date = cls._validate_params(type_, date)
        async with aiohttp.request(
            "GET", f"https://politicsandwar.com/data/{type_}/{type_}-{date}.csv.zip"
        ) as response:
            bytes_ = await response.read()
            return cls._parse_response_to_bytes(f"{type_}-{date}.csv", bytes_)

    @classmethod
    async def request_as_bytes(cls, type_: str, date: str) -> bytes:
        return await cls._request(type_, date)

    @classmethod
    async def request_as_str(cls, type_: str, date: str) -> str:
        bytes_ = await cls._request(type_, date)
        return bytes_.decode("utf-8")[1:]

    @classmethod
    async def request_as_csv(cls, type_: str, date: str) -> _reader:
        bytes_ = await cls._request(type_, date)
        return reader(bytes_.decode("utf-8")[1:].splitlines())

    @classmethod
    async def request_as_dicts(cls, type_: str, date: str) -> Sequence[Dict[str, Any]]:
        bytes_ = await cls._request(type_, date)
        reader = DictReader(bytes_.decode("utf-8")[1:].splitlines())
        return [i for i in reader]

    @classmethod
    async def request_as_dict(cls, type_: str, date: str) -> Dict[str, Dict[str, Any]]:
        dicts = await cls.request_as_dicts(type_, date)
        return {list(i.values())[0]: i for i in dicts}

    @classmethod
    async def request_as_data(cls, type_: str, date: str) -> Tuple[DumpData, ...]:
        data = await cls.request_as_dicts(type_, date)
        type__ = _TYPE_MAP[type_.lower()]
        return tuple(type__(i) for i in data)

    @classmethod
    async def request_as_data_dict(cls, type_: str, date: str) -> Dict[str, DumpData]:
        data = await cls.request_as_data(type_, date)
        return {str(int(i)): i for i in data}



"""
The MIT License (MIT)

Copyright (c) 2021-present Village

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, overload

from .data import *
from .legacy.core import Kit, async_pnwkit, pnwkit  # type: ignore
from .legacy.keys import set_bot_key, set_key
from .new import *

__all__ = (
    "set_key",
    "set_bot_key",
    "Kit",
    "async_pnwkit",
    "pnwkit",
    "alliance_query",
    "bankrec_query",
    "bbgame_query",
    "bbplayer_query",
    "bbteam_query",
    "bounty_query",
    "city_query",
    "color_query",
    "game_info_query",
    "me_query",
    "nation_query",
    "trade_query",
    "tradeprice_query",
    "treasure_query",
    "treaty_query",
    "war_query",
    "warattack_query",
    "bank_deposit_mutation",
    "bank_withdraw_mutation",
    "async_alliance_query",
    "async_bankrec_query",
    "async_bbgame_query",
    "async_bbplayer_query",
    "async_bbteam_query",
    "async_bounty_query",
    "async_city_query",
    "async_color_query",
    "async_game_info_query",
    "async_me_query",
    "async_nation_query",
    "async_trade_query",
    "async_tradeprice_query",
    "async_treasure_query",
    "async_treaty_query",
    "async_war_query",
    "async_warattack_query",
    "async_bank_deposit_mutation",
    "async_bank_withdraw_mutation",
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Coroutine,
        Literal,
        Mapping,
        MutableMapping,
        Tuple,
        TypeVar,
        Union,
    )

    from .legacy.data import *
    from .legacy.data import Data
    from .legacy.paginator import *

    D = TypeVar("D", bound=Data)

__version__ = "1.3.0"

@overload
def alliance_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Alliance, ...]: ...
@overload
def alliance_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Alliance]: ...
def alliance_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Alliance, ...], Paginator[Alliance]]: ...
@overload
def bankrec_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Bankrec, ...]: ...
@overload
def bankrec_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Bankrec]: ...
def bankrec_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Bankrec, ...], Paginator[Bankrec]]: ...
@overload
def bbgame_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[BBGame, ...]: ...
@overload
def bbgame_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[BBGame]: ...
def bbgame_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[BBGame, ...], Paginator[BBGame]]: ...
@overload
def bbplayer_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[BBPlayer, ...]: ...
@overload
def bbplayer_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[BBPlayer]: ...
def bbplayer_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[BBPlayer, ...], Paginator[BBPlayer]]: ...
@overload
def bbteam_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[BBTeam, ...]: ...
@overload
def bbteam_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[BBTeam]: ...
def bbteam_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[BBTeam, ...], Paginator[BBTeam]]: ...
@overload
def bounty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Bounty, ...]: ...
@overload
def bounty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Bounty]: ...
def bounty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Bounty, ...], Paginator[Bounty]]: ...
@overload
def city_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[City, ...]: ...
@overload
def city_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[City]: ...
def city_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[City, ...], Paginator[City]]: ...
def color_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
) -> Tuple[Color, ...]: ...
def game_info_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
) -> GameInfo: ...
def me_query(
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
) -> ApiKeyDetails: ...
@overload
def nation_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Nation, ...]: ...
@overload
def nation_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Nation]: ...
def nation_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Nation, ...], Paginator[Nation]]: ...
@overload
def trade_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Trade, ...]: ...
@overload
def trade_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Trade]: ...
def trade_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Trade, ...], Paginator[Trade]]: ...
@overload
def tradeprice_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Tradeprice, ...]: ...
@overload
def tradeprice_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Tradeprice]: ...
def tradeprice_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Tradeprice, ...], Paginator[Tradeprice]]: ...
def treasure_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
) -> Tuple[Treasure, ...]: ...
@overload
def treaty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[Treaty, ...]: ...
@overload
def treaty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Treaty]: ...
def treaty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[Treaty, ...], Paginator[Treaty]]: ...
@overload
def war_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[War, ...]: ...
@overload
def war_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[War]: ...
def war_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[War, ...], Paginator[War]]: ...
@overload
def warattack_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Tuple[WarAttack, ...]: ...
@overload
def warattack_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[WarAttack]: ...
def warattack_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Tuple[WarAttack, ...], Paginator[WarAttack]]: ...
def bank_deposit_mutation(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    **variables: Any,
) -> Bankrec: ...
def bank_withdraw_mutation(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    **variables: Any,
) -> Bankrec: ...
@overload
def async_alliance_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Alliance, ...]]: ...
@overload
def async_alliance_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Alliance]: ...
def async_alliance_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Alliance, ...]], Paginator[Alliance]]: ...
@overload
def async_bankrec_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Bankrec, ...]]: ...
@overload
def async_bankrec_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Bankrec]: ...
def async_bankrec_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Bankrec, ...]], Paginator[Bankrec]]: ...
@overload
def async_bbgame_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[BBGame, ...]]: ...
@overload
def async_bbgame_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[BBGame]: ...
def async_bbgame_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[BBGame, ...]], Paginator[BBGame]]: ...
@overload
def async_bbplayer_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[BBPlayer, ...]]: ...
@overload
def async_bbplayer_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[BBPlayer]: ...
def async_bbplayer_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[BBPlayer, ...]], Paginator[BBPlayer]]: ...
@overload
def async_bbteam_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[BBTeam, ...]]: ...
@overload
def async_bbteam_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[BBTeam]: ...
def async_bbteam_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[BBTeam, ...]], Paginator[BBTeam]]: ...
@overload
def async_bounty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Bounty, ...]]: ...
@overload
def async_bounty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Bounty]: ...
def async_bounty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Bounty, ...]], Paginator[Bounty]]: ...
@overload
def async_city_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[City, ...]]: ...
@overload
def async_city_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[City]: ...
def async_city_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[City, ...]], Paginator[City]]: ...
def async_color_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
) -> Coroutine[Any, Any, Tuple[Color, ...]]: ...
def async_game_info_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
) -> Coroutine[Any, Any, GameInfo]: ...
def async_me_query(
    params: MutableMapping[str, Any],
    arg: Union[str, Mapping[str, Any]],
    *args: Union[str, Mapping[str, Any]],
) -> Coroutine[Any, Any, ApiKeyDetails]: ...
@overload
def async_nation_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Nation, ...]]: ...
@overload
def async_nation_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Nation]: ...
def async_nation_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Nation, ...]], Paginator[Nation]]: ...
@overload
def async_trade_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Trade, ...]]: ...
@overload
def async_trade_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Trade]: ...
def async_trade_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Trade, ...]], Paginator[Trade]]: ...
@overload
def async_tradeprice_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Tradeprice, ...]]: ...
@overload
def async_tradeprice_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Tradeprice]: ...
def async_tradeprice_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Tradeprice, ...]], Paginator[Tradeprice]]: ...
def async_treasure_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
) -> Coroutine[Any, Any, Tuple[Treasure, ...]]: ...
@overload
def async_treaty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[Treaty, ...]]: ...
@overload
def async_treaty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[Treaty]: ...
def async_treaty_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[Treaty, ...]], Paginator[Treaty]]: ...
@overload
def async_war_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[War, ...]]: ...
@overload
def async_war_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[War]: ...
def async_war_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[War, ...]], Paginator[War]]: ...
@overload
def async_warattack_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[False] = ...,
    **variables: Any,
) -> Coroutine[Any, Any, Tuple[WarAttack, ...]]: ...
@overload
def async_warattack_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: Literal[True] = ...,
    **variables: Any,
) -> Paginator[WarAttack]: ...
def async_warattack_query(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    paginator: bool = ...,
    **variables: Any,
) -> Union[Coroutine[Any, Any, Tuple[WarAttack, ...]], Paginator[WarAttack]]: ...
def async_bank_deposit_mutation(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    **variables: Any,
) -> Coroutine[Any, Any, Bankrec]: ...
def async_bank_withdraw_mutation(
    params: MutableMapping[str, Any],
    *args: Union[str, Mapping[str, Any]],
    **variables: Any,
) -> Coroutine[Any, Any, Bankrec]: ...



from __future__ import annotations

import datetime
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, overload

from .keys import API_KEY, BOT_KEY

__all__ = ("KitBase",)

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Coroutine,
        List,
        Literal,
        Mapping,
        MutableMapping,
        Optional,
        Sequence,
        Tuple,
        Union,
    )

    from .data import *
    from .paginator import *

    SubQuery = Mapping[str, Union[str, "SubQuery", Sequence[Union[str, "SubQuery"]]]]


class KitBase(metaclass=ABCMeta):
    def __init__(
        self,
        api_key: Optional[str] = None,
        bot_key: Optional[str] = None,
        *,
        parse_int: Optional[Callable[[str], Any]] = None,
        parse_float: Optional[Callable[[str], Any]] = None,
    ) -> None:
        self.api_key: str = api_key or API_KEY
        self.bot_key: str = bot_key or BOT_KEY
        self.parse_int: Optional[Callable[[str], Any]] = parse_int
        self.parse_float: Optional[Callable[[str], Any]] = parse_float

    @property
    def graphql_url(self) -> str:
        return f"https://api.politicsandwar.com/graphql?api_key={self.api_key}"

    def set_key(self, api_key: str) -> None:
        """Sets the API key for this instance.

        Parameters
        ----------
        api_key : str
            A Politics and War API Key.
        """
        self.api_key: str = api_key

    def set_bot_key(self, bot_key: str, api_key: str) -> None:
        """Sets the bot key for this instance.

        Parameters
        ----------
        bot_key : str
            A Politics and War Verified Bot Key.
        api_key : str
            The API Key corresponding to the Bot Key.
        """
        self.bot_key: str = bot_key
        self.bot_key_api_key: str = api_key

    @classmethod
    def _format_sub_query(
        cls,
        query: SubQuery,
    ) -> str:
        key, query_string = list(query.items())[0]
        if not isinstance(query_string, str):
            query_arguments: List[Any] = []
            for value in query_string:
                if isinstance(value, str):
                    query_arguments.append(value)
                else:
                    query_arguments.append(cls._format_sub_query(value))
            query_string = " ".join(query_arguments)

        return f"{key}{{{query_string}}}"

    @classmethod
    def _format_query(
        cls,
        endpoint: str,
        params: MutableMapping[str, Any],
        query: Union[
            str,
            Sequence[Union[str, Mapping[str, Union[str, Mapping[str, Any]]]]],
        ],
        is_paginator: bool,
    ) -> str:
        if isinstance(params, str):
            params_string = params
        else:
            params_string = ",".join(
                name
                + ":"
                + (
                    (value if value.startswith("$") else f'"{value}"')
                    if isinstance(value, str)
                    else str(value).lower()
                    if isinstance(value, bool)
                    else (
                        "["
                        + ",".join(
                            (val if val.startswith("$") else f'"{val}"')
                            if isinstance(val, str)
                            else str(val).lower()
                            if isinstance(val, bool)
                            else str(val)  # type: ignore
                            for val in value  # type: ignore
                        )
                        + "]"
                    )
                    if isinstance(value, list)
                    else str(value)
                )
                for name, value in params.items()
            )
        if isinstance(query, str):
            query_string = query
        else:
            query_arguments: List[Any] = []
            for value in query:
                if isinstance(value, str):
                    query_arguments.append(value)
                else:
                    query_arguments.append(cls._format_sub_query(value))
            query_string = " ".join(query_arguments)
        if is_paginator:
            if params:
                return f"{{{endpoint}({params_string}){{paginatorInfo{{count currentPage firstItem hasMorePages lastItem lastPage perPage total}}data{{{query_string}}}}}}}"
            return f"{{{endpoint}{{paginatorInfo{{count currentPage firstItem hasMorePages lastItem lastPage perPage total}}data{{{query_string}}}}}}}"
        if params_string:
            return f"{{{endpoint}({params_string}){{{query_string}}}}}"
        return f"{{{endpoint}{{{query_string}}}}}"

    @classmethod
    def _format_variables(
        cls, variables: MutableMapping[str, Any]
    ) -> Tuple[MutableMapping[str, Any], str]:
        declarations: List[str] = []
        for name, value in variables.items():
            declarations.append(f"${name}:{cls._format_variable(value)}")
            if isinstance(value, datetime.datetime):
                variables[name] = str(value)
            elif isinstance(value, tuple):
                variables[name] = value[1]  # type: ignore
        return variables, ",".join(declarations)

    @classmethod
    def _format_variable(cls, value: Any) -> str:
        if isinstance(value, str):
            return "String"
        if isinstance(value, bool):
            return "Boolean"
        if isinstance(value, int):
            return "Int"
        if isinstance(value, float):
            return "Float"
        if isinstance(value, list):
            return f"[{cls._format_variable(value[0])}]"
        if isinstance(value, tuple):
            return value[0]  # type: ignore
        if isinstance(value, datetime.datetime):
            return "DateTime"
        raise ValueError(f"Unsupported variable type: {type(value)}")

    @overload
    @abstractmethod
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Alliance, ...]], Tuple[Alliance, ...]]:
        ...

    @overload
    @abstractmethod
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Alliance]:
        ...

    @abstractmethod
    def alliance_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Alliance, ...]],
        Tuple[Alliance, ...],
        Paginator[Alliance],
    ]:
        """Makes a query to the alliances endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Alliance, ...], Paginator[Alliance]]
            A tuple of Alliances representing the data retrived. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def bankrec_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Bankrec, ...]], Tuple[Bankrec, ...]]:
        ...

    @overload
    @abstractmethod
    def bankrec_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Bankrec]:
        ...

    @abstractmethod
    def bankrec_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Bankrec, ...]],
        Tuple[Bankrec, ...],
        Paginator[Bankrec],
    ]:
        """Makes a query to the bankrecs endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Bankrec, ...], Paginator[Bankrec]]
            A tuple of Bankrecs representing the data retrived. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def baseball_game_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[BBGame, ...]], Tuple[BBGame, ...]]:
        ...

    @overload
    @abstractmethod
    def baseball_game_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[BBGame]:
        ...

    @abstractmethod
    def baseball_game_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[BBGame, ...]],
        Tuple[BBGame, ...],
        Paginator[BBGame],
    ]:
        """Makes a query to the baseball_games endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[BBGame, ...], Paginator[BBGame]]
            A tuple of BBGames representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def baseball_player_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[BBPlayer, ...]], Tuple[BBPlayer, ...]]:
        ...

    @overload
    @abstractmethod
    def baseball_player_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[BBPlayer]:
        ...

    @abstractmethod
    def baseball_player_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[BBPlayer, ...]],
        Tuple[BBPlayer, ...],
        Paginator[BBPlayer],
    ]:
        """Makes a query to the baseball_players endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[BBPlayer, ...], Paginator[BBPlayer]]
            A tuple of BBPlayers representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def baseball_team_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[BBTeam, ...]], Tuple[BBTeam, ...]]:
        ...

    @overload
    @abstractmethod
    def baseball_team_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[BBTeam]:
        ...

    @abstractmethod
    def baseball_team_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[BBTeam, ...]],
        Tuple[BBTeam, ...],
        Paginator[BBTeam],
    ]:
        """Makes a query to the baseball_teams endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[BBTeam, ...], Paginator[BBTeam]]
            A tuple of BBTeams representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def bounty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Bounty, ...]], Tuple[Bounty, ...]]:
        ...

    @overload
    @abstractmethod
    def bounty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Bounty]:
        ...

    @abstractmethod
    def bounty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Bounty, ...]],
        Tuple[Bounty, ...],
        Paginator[Bounty],
    ]:
        """Makes a query to the bounties endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Bounty, ...], Paginator[Bounty]]
            A tuple of Bounties representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def city_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[City, ...]], Tuple[City, ...]]:
        ...

    @overload
    @abstractmethod
    def city_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[City]:
        ...

    @abstractmethod
    def city_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[City, ...]],
        Tuple[City, ...],
        Paginator[City],
    ]:
        """Makes a query to the cities endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[City, ...], Paginator[City]]
            A tuple of Cities representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @abstractmethod
    def color_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
    ) -> Union[Coroutine[Any, Any, Tuple[Color, ...]], Tuple[Color, ...]]:
        """Makes a query to the colors endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.

        Returns
        -------
        Tuple[Color, ...]
            A tuple of Colors representing the data retrived.
        """
        ...

    @abstractmethod
    def game_info_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
    ) -> Union[Coroutine[Any, Any, GameInfo], GameInfo]:
        """Makes a query to the game_info endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.

        Returns
        -------
        GameInfo
            The GameInfo object representing the data retrieved.
        """
        ...

    @abstractmethod
    def me_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
    ) -> Union[Coroutine[Any, Any, ApiKeyDetails], ApiKeyDetails]:
        """Makes a query to the me endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.

        Returns
        -------
        ApiKeyDetails
            The ApiKeyDetails object representing the data retrieved.
        """
        ...

    @overload
    @abstractmethod
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Nation, ...]], Tuple[Nation, ...]]:
        ...

    @overload
    @abstractmethod
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Nation]:
        ...

    @abstractmethod
    def nation_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Nation, ...]],
        Tuple[Nation, ...],
        Paginator[Nation],
    ]:
        """Makes a query to the nations endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Nation, ...], Paginator[Nation]]
            A tuple of Nations representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def trade_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Trade, ...]], Tuple[Trade, ...]]:
        ...

    @overload
    @abstractmethod
    def trade_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Trade]:
        ...

    @abstractmethod
    def trade_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Trade, ...]],
        Tuple[Trade, ...],
        Paginator[Trade],
    ]:
        """Makes a query to the trades endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Trade, ...], Paginator[Trade]]
            A tuple of Trades representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def tradeprice_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Tradeprice, ...]], Tuple[Tradeprice, ...]]:
        ...

    @overload
    @abstractmethod
    def tradeprice_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Tradeprice]:
        ...

    @abstractmethod
    def tradeprice_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Tradeprice, ...]],
        Tuple[Tradeprice, ...],
        Paginator[Tradeprice],
    ]:
        """Makes a query to the tradeprices endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Tradeprice, ...], Paginator[Tradeprice]]
            A tuple of Tradeprices representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @abstractmethod
    def treasure_query(
        self,
        params: MutableMapping[str, Any],
        arg: Union[str, Mapping[str, Any]],
        *args: Union[str, Mapping[str, Any]],
    ) -> Union[Coroutine[Any, Any, Tuple[Treasure, ...]], Tuple[Treasure, ...]]:
        """Makes a query to the treasures endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.

        Returns
        -------
        Tuple[Treasure, ...]
            A tuple of Treasures representing the data retrived.
        """
        ...

    @overload
    @abstractmethod
    def treaty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[Treaty, ...]], Tuple[Treaty, ...]]:
        ...

    @overload
    @abstractmethod
    def treaty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[Treaty]:
        ...

    @abstractmethod
    def treaty_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[Treaty, ...]],
        Tuple[Treaty, ...],
        Paginator[Treaty],
    ]:
        """Makes a query to the treaties endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[Treaty, ...], Paginator[Treaty]]
            A tuple of Treaties representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def war_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[War, ...]], Tuple[War, ...]]:
        ...

    @overload
    @abstractmethod
    def war_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[War]:
        ...

    @abstractmethod
    def war_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[War, ...]], Tuple[War, ...], Paginator[War]]:
        """Makes a query to the wars endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[War, ...], Paginator[War]]
            A tuple of Wars representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @overload
    @abstractmethod
    def warattack_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[False] = ...,
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Tuple[WarAttack, ...]], Tuple[WarAttack, ...]]:
        ...

    @overload
    @abstractmethod
    def warattack_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: Literal[True] = ...,
        **variables: Any,
    ) -> Paginator[WarAttack]:
        ...

    @abstractmethod
    def warattack_query(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        paginator: bool = False,
        **variables: Any,
    ) -> Union[
        Coroutine[Any, Any, Tuple[WarAttack, ...]],
        Tuple[WarAttack, ...],
        Paginator[WarAttack],
    ]:
        """Makes a query to the warattacks endpoint.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass in the query to filter results.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        paginator : bool
            Whether to return the result as a Paginator, by default False.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Union[Tuple[WarAttack, ...], Paginator[WarAttack]]
            A tuple of WarAttacks representing the data retrieved. If ``paginator`` is True then will return an Paginator.
        """
        ...

    @abstractmethod
    def bank_deposit_mutation(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Bankrec], Bankrec]:
        """Executes the bank deposit mutation.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass into the mutation.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Bankrec
            The Bankrec object representing the transaction completed.
        """
        ...

    @abstractmethod
    def bank_withdraw_mutation(
        self,
        params: MutableMapping[str, Any],
        *args: Union[str, Mapping[str, Any]],
        **variables: Any,
    ) -> Union[Coroutine[Any, Any, Bankrec], Bankrec]:
        """Executes the bank withdraw mutation.

        Parameters
        ----------
        params : Union[str, Mapping[str, Any]]
            A string, dict, or other mapping of the parameters to pass into the mutation.
        *args: Union[str, Mapping[str, Any]]
            Will be concatenated to form the portion of the query string containing the attributes to fetch.
        **variables: Any
            The variables to pass in with the query.

        Returns
        -------
        Bankrec
            The Bankrec object representing the transaction completed.
        """
        ...
